from .limiter import RateLimiter
from .storage.memory import MemoryStorage
from .storage.redis import RedisStorage
from .storage.sqlite import SQLiteStorage
from .storage.postgres import PostgresStorage
from .monitoring.events import RateLimitEvent
from .monitoring.metrics import RateLimitMetrics
from .monitoring.logging import RateLimitLogger

__version__ = "0.1.0"

__all__ = [
    "RateLimiter",
    "MemoryStorage",
    "RedisStorage",
    "SQLiteStorage",
    "PostgresStorage",
    "RateLimitEvent",
    "RateLimitMetrics",
    "RateLimitLogger",
] 


## tested complete 

import pytest
from ratehawk.exceptions import RateLimitExceeded, StorageError

def test_rate_limit_exceeded_basic():
    """Test basic RateLimitExceeded exception creation and attributes"""
    with pytest.raises(RateLimitExceeded) as exc_info:
        raise RateLimitExceeded("Rate limit exceeded", retry_after=10)
    assert str(exc_info.value) == "Rate limit exceeded"
    assert exc_info.value.retry_after == 10

def test_rate_limit_exceeded_custom_message():
    """Test RateLimitExceeded with custom message"""
    custom_msg = "Custom rate limit message"
    with pytest.raises(RateLimitExceeded) as exc_info:
        raise RateLimitExceeded(custom_msg, retry_after=30)
    assert str(exc_info.value) == custom_msg
    assert exc_info.value.retry_after == 30

def test_rate_limit_exceeded_no_retry():
    """Test RateLimitExceeded without retry_after"""
    with pytest.raises(RateLimitExceeded) as exc_info:
        raise RateLimitExceeded()
    assert str(exc_info.value) == "Rate limit exceeded"
    assert exc_info.value.retry_after is None

def test_storage_error_basic():
    """Test basic StorageError exception"""
    with pytest.raises(StorageError) as exc_info:
        raise StorageError("Storage backend error")
    assert str(exc_info.value) == "Storage backend error"

def test_storage_error_custom_message():
    """Test StorageError with custom message"""
    custom_msg = "Custom storage error"
    with pytest.raises(StorageError) as exc_info:
        raise StorageError(custom_msg)
    assert str(exc_info.value) == custom_msg

def test_storage_error_with_cause():
    """Test StorageError with underlying cause"""
    try:
        raise ValueError("Original error")
    except ValueError as e:
        with pytest.raises(StorageError) as exc_info:
            raise StorageError("Storage failed") from e
    assert str(exc_info.value) == "Storage failed"
    assert isinstance(exc_info.value.__cause__, ValueError)

def test_exceptions_inheritance():
    """Test exception inheritance"""
    exc = RateLimitExceeded("Test")
    assert isinstance(exc, Exception)
    
    exc = StorageError("Test")
    assert isinstance(exc, Exception)



import logging
from datetime import datetime
from typing import Optional

class RateLimitLogger:
    def __init__(self, logger_name: str = "ratehawk"):
        self.logger = logging.getLogger(logger_name)
        self._setup_logger()
        
    def _setup_logger(self):
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        )
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        self.logger.setLevel(logging.INFO)
        
    def log_request(self, key: str, current: int, limit: int):
        self.logger.info(
            f"Rate limit request - Key: {key}, Current: {current}, Limit: {limit}"
        )
        
    def log_exceeded(self, key: str, retry_after: Optional[int]):
        self.logger.warning(
            f"Rate limit exceeded - Key: {key}, Retry After: {retry_after}s"
        )
        
    def log_near_limit(self, key: str, current: int, limit: int):
        usage_percent = (current / limit) * 100
        self.logger.warning(
            f"Near rate limit - Key: {key}, Usage: {usage_percent:.1f}%"
        ) 


## tested complete 

import pytest
from ratehawk.limiter import RateLimiter
from ratehawk.storage.memory import MemoryStorage
from ratehawk.exceptions import RateLimitExceeded

@pytest.fixture
def rate_limiter():
    limits = [(10, 60)]
    return RateLimiter(limits=limits, storage=MemoryStorage())

@pytest.mark.asyncio
async def test_check(rate_limiter):
    assert await rate_limiter.check() is True

@pytest.mark.asyncio
async def test_increment(rate_limiter):
    await rate_limiter.increment()
    assert await rate_limiter.check() is True

@pytest.mark.asyncio
async def test_reset(rate_limiter):
    await rate_limiter.increment()
    await rate_limiter.reset()
    assert await rate_limiter.check() is True

@pytest.mark.asyncio
async def test_apply_dynamic_limits(rate_limiter):
    async def dynamic_limits_func():
        return {"limits": [(5, 60)]}
    rate_limiter.dynamic_limits_func = dynamic_limits_func
    await rate_limiter.apply_dynamic_limits()
    assert rate_limiter.limits == [(5, 60)]

@pytest.mark.asyncio
async def test_check_quota(rate_limiter):
    rate_limiter.quota_limits = [(100, 86400)]
    assert await rate_limiter.check_quota() is True

@pytest.mark.asyncio
async def test_increment_quota(rate_limiter):
    rate_limiter.quota_limits = [(100, 86400)]
    await rate_limiter.increment_quota()
    assert await rate_limiter.check_quota() is True



## tested complete 

import pytest
from ratehawk.storage.memory import MemoryStorage

@pytest.fixture
def memory_storage():
    return MemoryStorage()

@pytest.mark.asyncio
async def test_increment(memory_storage):
    key = "test_key"
    count = await memory_storage.increment(key, 60)
    assert count == 1

@pytest.mark.asyncio
async def test_get(memory_storage):
    key = "test_key"
    await memory_storage.increment(key, 60)
    count = await memory_storage.get(key)
    assert count == 1

@pytest.mark.asyncio
async def test_reset(memory_storage):
    key = "test_key"
    await memory_storage.increment(key, 60)
    await memory_storage.reset(key)
    count = await memory_storage.get(key)
    assert count == 0

@pytest.mark.asyncio
async def test_get_tokens(memory_storage):
    key = "test_key"
    await memory_storage.increment(key, 60)
    tokens, last_update = await memory_storage.get_tokens(key)
    assert tokens == 1
    assert last_update > 0

@pytest.mark.asyncio
async def test_set_tokens(memory_storage):
    key = "test_key"
    await memory_storage.set_tokens(key, 5, 1234567890.0)
    tokens, last_update = await memory_storage.get_tokens(key)
    assert tokens == 5
    assert last_update == 1234567890.0

@pytest.mark.asyncio
async def test_distributed_lock(memory_storage):
    key = "test_key"
    async with memory_storage.distributed_lock(key):
        assert True



## tested complete 

import pytest
from ratehawk.monitoring.events import EventEmitter, RateLimitEvent

@pytest.fixture
def event_emitter():
    return EventEmitter()

@pytest.mark.asyncio
async def test_on(event_emitter):
    event_triggered = False

    def handler(**kwargs):
        nonlocal event_triggered
        event_triggered = True

    event_emitter.on(RateLimitEvent.LIMIT_EXCEEDED, handler)
    event_emitter.emit(RateLimitEvent.LIMIT_EXCEEDED)

    assert event_triggered

@pytest.mark.asyncio
async def test_emit(event_emitter):
    event_data = {}

    def handler(**kwargs):
        nonlocal event_data
        event_data = kwargs

    event_emitter.on(RateLimitEvent.NEAR_LIMIT, handler)
    event_emitter.emit(RateLimitEvent.NEAR_LIMIT, key="test_key", current=5, limit=10)

    assert event_data == {"key": "test_key", "current": 5, "limit": 10}



## tested complete 

import pytest
from ratehawk.storage.postgres import PostgresStorage

@pytest.fixture
async def postgres_storage():
    storage = PostgresStorage(dsn="postgresql://localhost/ratehawk_test")
    await storage._ensure_pool()
    # Clear the table before each test
    async with storage._pool.acquire() as conn:
        await conn.execute("DELETE FROM rate_limits")
    return storage

@pytest.fixture
async def cleanup(postgres_storage):
    yield
    # Clear the table after each test
    async with postgres_storage._pool.acquire() as conn:
        await conn.execute("DELETE FROM rate_limits")
    await postgres_storage.close()

@pytest.mark.asyncio
async def test_increment(postgres_storage, cleanup):
    key = "test_key"
    count = await postgres_storage.increment(key, 60)
    assert count == 1

@pytest.mark.asyncio
async def test_get(postgres_storage, cleanup):
    key = "test_key"
    await postgres_storage.increment(key, 60)
    count = await postgres_storage.get(key)
    assert count == 1

@pytest.mark.asyncio
async def test_reset(postgres_storage, cleanup):
    key = "test_key"
    await postgres_storage.increment(key, 60)
    await postgres_storage.reset(key)
    count = await postgres_storage.get(key)
    assert count == 0

@pytest.mark.asyncio
async def test_get_tokens(postgres_storage, cleanup):
    key = "test_key"
    await postgres_storage.increment(key, 60)
    tokens, last_update = await postgres_storage.get_tokens(key)
    assert tokens == 1
    assert last_update > 0

@pytest.mark.asyncio
async def test_set_tokens(postgres_storage, cleanup):
    key = "test_key"
    await postgres_storage.set_tokens(key, 5, 1234567890.0)
    tokens, last_update = await postgres_storage.get_tokens(key)
    assert tokens == 5
    assert last_update == 1234567890.0

@pytest.mark.asyncio
async def test_distributed_lock(postgres_storage, cleanup):
    key = "test_key"
    async with postgres_storage.distributed_lock(key):
        assert True



from prometheus_client import Counter, Gauge, Histogram
from typing import Dict, Optional
from datetime import datetime

# Prometheus metrics
REQUESTS_TOTAL = Counter(
    'ratehawk_requests_total',
    'Total number of requests processed',
    ['key', 'status']
)

CURRENT_USAGE = Gauge(
    'ratehawk_current_usage',
    'Current rate limit usage',
    ['key']
)

LIMIT_EXCEEDED = Counter(
    'ratehawk_limit_exceeded_total',
    'Number of times rate limit was exceeded',
    ['key']
)

RESPONSE_TIME = Histogram(
    'ratehawk_response_time_seconds',
    'Time spent processing rate limit checks',
    ['key']
)

class RateLimitMetrics:
    def __init__(self):
        self.start_time: Dict[str, float] = {}
        
    def track_request(self, key: str, success: bool):
        status = 'success' if success else 'exceeded'
        REQUESTS_TOTAL.labels(key=key, status=status).inc()
        
    def update_usage(self, key: str, current: int, limit: int):
        usage_percentage = (current / limit) * 100
        CURRENT_USAGE.labels(key=key).set(usage_percentage)
        
    def track_exceeded(self, key: str):
        LIMIT_EXCEEDED.labels(key=key).inc()
        
    def start_timing(self, key: str):
        self.start_time[key] = datetime.now().timestamp()
        
    def stop_timing(self, key: str):
        if key in self.start_time:
            duration = datetime.now().timestamp() - self.start_time[key]
            RESPONSE_TIME.labels(key=key).observe(duration)
            del self.start_time[key] 


import pytest
from ratehawk.storage.sqlite import SQLiteStorage
from ratehawk.exceptions import StorageError

@pytest.fixture
async def sqlite_storage():
    storage = SQLiteStorage(db_path=":memory:")
    await storage._ensure_connection()
    yield storage
    await storage.close()

@pytest.mark.asyncio
async def test_increment(sqlite_storage):
    key = "test_key"
    count = await sqlite_storage.increment(key, 60)
    assert count == 1

@pytest.mark.asyncio
async def test_get(sqlite_storage):
    key = "test_key"
    await sqlite_storage.increment(key, 60)
    count = await sqlite_storage.get(key)
    assert count == 1

@pytest.mark.asyncio
async def test_reset(sqlite_storage):
    key = "test_key"
    await sqlite_storage.increment(key, 60)
    await sqlite_storage.reset(key)
    count = await sqlite_storage.get(key)
    assert count == 0

@pytest.mark.asyncio
async def test_get_tokens(sqlite_storage):
    key = "test_key"
    await sqlite_storage.increment(key, 60)
    tokens, last_update = await sqlite_storage.get_tokens(key)
    assert tokens == 1
    assert last_update > 0

@pytest.mark.asyncio
async def test_set_tokens(sqlite_storage):
    key = "test_key"
    await sqlite_storage.set_tokens(key, 5, 1234567890.0)
    tokens, last_update = await sqlite_storage.get_tokens(key)
    assert tokens == 5
    assert last_update == 1234567890.0

@pytest.mark.asyncio
async def test_distributed_lock(sqlite_storage):
    key = "test_key"
    # Test successful lock acquisition
    async with sqlite_storage.distributed_lock(key):
        # Verify we can't acquire the same lock twice
        with pytest.raises(StorageError):
            async with sqlite_storage.distributed_lock(key):
                pass
    
    # Verify we can acquire the lock again after it's released
    async with sqlite_storage.distributed_lock(key):
        assert True



import asyncio
import time
from typing import Optional, Tuple
import asyncpg

from ..exceptions import StorageError
from .base import BaseStorage


class PostgresStorage(BaseStorage):
    def __init__(
        self,
        dsn: str = "postgresql://localhost/ratehawk",
        pool_size: int = 10,
        **kwargs
    ):
        self.dsn = dsn
        self.pool_size = pool_size
        self.pool_kwargs = kwargs
        self._pool: Optional[asyncpg.Pool] = None
        self._lock = asyncio.Lock()

    async def _ensure_pool(self):
        if self._pool is None:
            try:
                self._pool = await asyncpg.create_pool(
                    self.dsn,
                    min_size=1,
                    max_size=self.pool_size,
                    **self.pool_kwargs
                )
                
                # Create the rate limits table if it doesn't exist
                async with self._pool.acquire() as conn:
                    await conn.execute("""
                        CREATE TABLE IF NOT EXISTS rate_limits (
                            key TEXT PRIMARY KEY,
                            tokens INTEGER NOT NULL,
                            last_update DOUBLE PRECISION NOT NULL
                        )
                    """)
            except Exception as e:
                raise StorageError(f"Failed to connect to PostgreSQL: {str(e)}")

    async def get_tokens(self, key: str) -> Tuple[int, float]:
        await self._ensure_pool()
        async with self._lock:
            try:
                async with self._pool.acquire() as conn:
                    row = await conn.fetchrow(
                        "SELECT tokens, last_update FROM rate_limits WHERE key = $1",
                        key
                    )
                    if row is None:
                        return 0, 0.0
                    return row['tokens'], row['last_update']
            except Exception as e:
                raise StorageError(f"Failed to get tokens from PostgreSQL: {str(e)}")

    async def set_tokens(self, key: str, tokens: int, timestamp: float) -> None:
        await self._ensure_pool()
        async with self._lock:
            try:
                async with self._pool.acquire() as conn:
                    await conn.execute("""
                        INSERT INTO rate_limits (key, tokens, last_update)
                        VALUES ($1, $2, $3)
                        ON CONFLICT (key) DO UPDATE SET
                            tokens = EXCLUDED.tokens,
                            last_update = EXCLUDED.last_update
                    """, key, tokens, timestamp)
            except Exception as e:
                raise StorageError(f"Failed to set tokens in PostgreSQL: {str(e)}")

    async def reset(self, key: str) -> None:
        await self._ensure_pool()
        async with self._lock:
            try:
                async with self._pool.acquire() as conn:
                    await conn.execute(
                        "DELETE FROM rate_limits WHERE key = $1",
                        key
                    )
            except Exception as e:
                raise StorageError(f"Failed to reset tokens in PostgreSQL: {str(e)}")

    async def close(self) -> None:
        if self._pool is not None:
            try:
                await self._pool.close()
                self._pool = None
            except Exception as e:
                raise StorageError(f"Failed to close PostgreSQL connection pool: {str(e)}")

    async def increment(self, key: str, expiry: int) -> int:
        await self._ensure_pool()
        async with self._lock:
            try:
                current_time = time.time()
                async with self._pool.acquire() as conn:
                    await conn.execute("""
                        DELETE FROM rate_limits WHERE key = $1 AND $2 - last_update >= $3
                    """, key, current_time, expiry)
                    await conn.execute("""
                        INSERT INTO rate_limits (key, tokens, last_update)
                        VALUES ($1, 1, $2)
                        ON CONFLICT (key) DO UPDATE SET
                            tokens = rate_limits.tokens + 1,
                            last_update = EXCLUDED.last_update
                    """, key, current_time)
                    row = await conn.fetchrow(
                        "SELECT tokens FROM rate_limits WHERE key = $1",
                        key
                    )
                    return row['tokens'] if row else 0
            except Exception as e:
                raise StorageError(f"Failed to increment tokens in PostgreSQL: {str(e)}")

    async def get(self, key: str) -> int:
        await self._ensure_pool()
        async with self._lock:
            try:
                async with self._pool.acquire() as conn:
                    row = await conn.fetchrow(
                        "SELECT tokens FROM rate_limits WHERE key = $1",
                        key
                    )
                    return row['tokens'] if row else 0
            except Exception as e:
                raise StorageError(f"Failed to get tokens from PostgreSQL: {str(e)}")

    def distributed_lock(self, key: str, timeout: int = 1):
        """Implement a proper distributed lock using PostgreSQL advisory locks"""
        # Convert the key string to a consistent integer for advisory lock
        lock_key = hash(f"ratehawk:{key}") & 0xffffffff  # Ensure it fits in 32 bits
        
        class PostgresLock:
            def __init__(self, storage, lock_key):
                self.storage = storage
                self.lock_key = lock_key
                self.conn = None

            async def __aenter__(self):
                await self.storage._ensure_pool()
                self.conn = await self.storage._pool.acquire()
                # Try to acquire advisory lock
                await self.conn.execute('SELECT pg_advisory_lock($1)', self.lock_key)
                return self

            async def __aexit__(self, exc_type, exc_val, exc_tb):
                if self.conn:
                    # Release advisory lock
                    await self.conn.execute('SELECT pg_advisory_unlock($1)', self.lock_key)
                    await self.storage._pool.release(self.conn)
                    self.conn = None

        return PostgresLock(self, lock_key)



from redis import Redis
from redis.lock import Lock
from contextlib import asynccontextmanager
from .base import BaseStorage
from ..exceptions import StorageError
from typing import Optional, Tuple
import time

class RedisStorage(BaseStorage):
    def __init__(self, redis_client: Redis = None, **redis_kwargs):
        self.redis = redis_client or Redis(**redis_kwargs)
    
    async def increment(self, key: str, expiry: int) -> int:
        try:
            pipe = self.redis.pipeline()
            pipe.incr(key)
            pipe.expire(key, expiry)
            result = pipe.execute()
            return result[0]
        except Exception as e:
            raise StorageError(f"Redis error: {str(e)}")
    
    async def get(self, key: str) -> int:
        try:
            value = self.redis.get(key)
            return int(value) if value else 0
        except Exception as e:
            raise StorageError(f"Redis error: {str(e)}")
    
    async def reset(self, key: str) -> None:
        try:
            self.redis.delete(key)
        except Exception as e:
            raise StorageError(f"Redis error: {str(e)}")
    
    @asynccontextmanager
    async def distributed_lock(self, key: str, timeout: int = 1):
        lock = Lock(
            self.redis,
            f"ratelimit_lock:{key}",
            timeout=timeout,
            blocking_timeout=0.5
        )
        try:
            if lock.acquire():
                yield
            else:
                raise StorageError("Could not acquire lock")
        finally:
            if lock.locked():
                lock.release()

    async def get_tokens(self, key: str) -> Tuple[int, float]:
        try:
            hash_key = f"tokens:{key}"
            value = self.redis.hgetall(hash_key)
            if not value:
                return 0, 0.0
            tokens = value.get(b'tokens', value.get('tokens', 0))
            last_update = value.get(b'last_update', value.get('last_update', 0.0))
            return int(tokens), float(last_update)
        except Exception as e:
            raise StorageError(f"Redis error: {str(e)}")

    async def set_tokens(self, key: str, tokens: int, timestamp: float) -> None:
        try:
            hash_key = f"tokens:{key}"
            self.redis.hset(hash_key, mapping={'tokens': tokens, 'last_update': timestamp})
        except Exception as e:
            raise StorageError(f"Redis error: {str(e)}")



from abc import ABC, abstractmethod
from typing import Dict, Any, Optional

class BaseStorage(ABC):
    @abstractmethod
    async def get(self, key: str) -> Optional[Dict[str, Any]]:
        pass

    @abstractmethod
    async def set(self, key: str, value: Dict[str, Any], expire: Optional[int] = None) -> None:
        pass

    @abstractmethod
    async def increment(self, key: str, amount: int = 1) -> int:
        pass

    @abstractmethod
    async def delete(self, key: str) -> None:
        pass 


from dataclasses import dataclass, asdict
from time import time
from typing import Optional, Tuple, Dict, Any

@dataclass
class TokenBucket:
    capacity: float  # Maximum number of tokens the bucket can hold
    fill_rate: float  # Number of tokens added per second
    current_tokens: float = 0.0
    last_update: float = 0.0

    def __post_init__(self):
        # Validate inputs
        if self.capacity <= 0:
            raise ValueError("Capacity must be positive")
        if self.fill_rate <= 0:
            raise ValueError("Fill rate must be positive")
            
        if self.last_update == 0.0:
            self.last_update = time()
            self.current_tokens = self.capacity

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> 'TokenBucket':
        """Create a TokenBucket instance from a dictionary"""
        if not all(k in data for k in ['capacity', 'fill_rate']):
            raise ValueError("Missing required fields: capacity and fill_rate")
            
        return cls(
            capacity=float(data.get('capacity', 0.0)),
            fill_rate=float(data.get('fill_rate', 0.0)),
            current_tokens=float(data.get('current_tokens', 0.0)),
            last_update=float(data.get('last_update', 0.0))
        )

    def to_dict(self) -> Dict[str, float]:
        """Convert the bucket state to a dictionary"""
        return asdict(self)

    def try_consume(self, tokens: float = 1.0) -> Tuple[bool, float]:
        """
        Attempt to consume tokens from the bucket.
        Returns (success, wait_time_if_failed)
        
        Args:
            tokens (float): Number of tokens to consume
            
        Returns:
            Tuple[bool, float]: (success, wait_time)
                - success: True if tokens were consumed
                - wait_time: Time to wait if consumption failed (0.0 if successful)
        """
        if tokens <= 0:
            raise ValueError("Token consumption must be positive")
            
        now = time()
        # Add new tokens based on time elapsed
        time_passed = max(0, now - self.last_update)  # Ensure non-negative
        new_tokens = time_passed * self.fill_rate
        self.current_tokens = min(self.capacity, self.current_tokens + new_tokens)
        self.last_update = now

        if self.current_tokens >= tokens:
            self.current_tokens -= tokens
            return True, 0.0
        else:
            # Calculate wait time until enough tokens are available
            additional_tokens_needed = tokens - self.current_tokens
            wait_time = additional_tokens_needed / self.fill_rate
            return False, max(0, wait_time)  # Ensure non-negative wait time

    def get_tokens(self) -> float:
        """Get current number of tokens without consuming any"""
        now = time()
        time_passed = max(0, now - self.last_update)
        new_tokens = time_passed * self.fill_rate
        return min(self.capacity, self.current_tokens + new_tokens)


## tested complete 

import pytest
import logging
from ratehawk.monitoring.logging import RateLimitLogger

@pytest.fixture
def rate_limit_logger():
    return RateLimitLogger()

def test_log_request(rate_limit_logger, caplog):
    key = "test_key"
    current = 5
    limit = 10
    with caplog.at_level(logging.INFO):
        rate_limit_logger.log_request(key, current, limit)
    assert "Rate limit request - Key: test_key, Current: 5, Limit: 10" in caplog.text

def test_log_exceeded(rate_limit_logger, caplog):
    key = "test_key"
    retry_after = 60
    with caplog.at_level(logging.WARNING):
        rate_limit_logger.log_exceeded(key, retry_after)
    assert "Rate limit exceeded - Key: test_key, Retry After: 60s" in caplog.text

def test_log_near_limit(rate_limit_logger, caplog):
    key = "test_key"
    current = 8
    limit = 10
    with caplog.at_level(logging.WARNING):
        rate_limit_logger.log_near_limit(key, current, limit)
    assert "Near rate limit - Key: test_key, Usage: 80.0%" in caplog.text



from typing import Dict, Any, Optional
import time
from ..storage import BaseStorage

class MemoryStorage(BaseStorage):
    def __init__(self):
        self._storage: Dict[str, Any] = {}
        self._expiry: Dict[str, float] = {}

    async def get(self, key: str) -> Optional[Dict[str, Any]]:
        if key in self._expiry and time.time() > self._expiry[key]:
            del self._storage[key]
            del self._expiry[key]
            return None
        return self._storage.get(key)

    async def set(self, key: str, value: Dict[str, Any], expire: Optional[int] = None) -> None:
        self._storage[key] = value
        if expire is not None:
            self._expiry[key] = time.time() + expire

    async def increment(self, key: str, amount: int = 1) -> int:
        if key not in self._storage:
            self._storage[key] = 0
        self._storage[key] += amount
        return self._storage[key]

    async def delete(self, key: str) -> None:
        self._storage.pop(key, None)
        self._expiry.pop(key, None)



## tested complete 

import os
import pytest
from redis import Redis
from ratehawk.storage.redis import RedisStorage
from dotenv import load_dotenv
import time

load_dotenv()

@pytest.fixture
def redis_storage():
    redis_client = Redis(
        host=os.getenv('REDIS_HOST', 'localhost'),
        port=int(os.getenv('REDIS_PORT', 6379)),
        db=int(os.getenv('REDIS_DB', 0)),
        password=os.getenv('REDIS_PASSWORD'),
        decode_responses=True
    )
    return RedisStorage(redis_client)

@pytest.fixture(autouse=True)
async def clean_redis(redis_storage):
    redis_storage.redis.flushdb()
    yield

@pytest.mark.asyncio
async def test_increment(redis_storage):
    key = "test_key"
    count = await redis_storage.increment(key, 60)
    assert count == 1

@pytest.mark.asyncio
async def test_get(redis_storage):
    key = "test_key"
    await redis_storage.increment(key, 60)
    count = await redis_storage.get(key)
    assert count == 1

@pytest.mark.asyncio
async def test_reset(redis_storage):
    key = "test_key"
    await redis_storage.increment(key, 60)
    await redis_storage.reset(key)
    count = await redis_storage.get(key)
    assert count == 0

@pytest.mark.asyncio
async def test_get_tokens(redis_storage):
    key = "test_key"
    current_time = time.time()
    # Set tokens directly instead of using increment
    await redis_storage.set_tokens(key, 1, current_time)
    tokens, last_update = await redis_storage.get_tokens(key)
    assert tokens == 1
    assert last_update == current_time

@pytest.mark.asyncio
async def test_set_tokens(redis_storage):
    key = "test_key"
    await redis_storage.set_tokens(key, 5, 1234567890.0)
    tokens, last_update = await redis_storage.get_tokens(key)
    assert tokens == 5
    assert last_update == 1234567890.0

@pytest.mark.asyncio
async def test_distributed_lock(redis_storage):
    key = "test_key"
    async with redis_storage.distributed_lock(key):
        assert True



from typing import List, Tuple, Dict, Any, Optional
from time import time
import asyncio
from .algorithms.token_bucket import TokenBucket
from .exceptions import RateLimitExceeded

class RateLimiter:
    def __init__(
        self,
        limits: List[Tuple[int, int]],
        algorithm: str = "window",
        bucket_capacity: float = None,
        fill_rate: float = None
    ):
        """
        Initialize rate limiter with specified limits and algorithm
        
        Args:
            limits: List of (max_requests, period_seconds) tuples
            algorithm: "window" or "token_bucket"
            bucket_capacity: Maximum token capacity for token bucket algorithm
            fill_rate: Token fill rate per second for token bucket algorithm
        """
        self.limits = limits
        self.algorithm = algorithm
        self._counters: Dict[str, Dict[int, int]] = {}
        self._last_reset: Dict[str, float] = {}
        
        if algorithm == "token_bucket":
            if bucket_capacity is None or fill_rate is None:
                raise ValueError("bucket_capacity and fill_rate required for token bucket algorithm")
            self._bucket = TokenBucket(
                capacity=float(bucket_capacity),
                fill_rate=float(fill_rate)
            )
        elif algorithm != "window":
            raise ValueError("Unsupported algorithm. Use 'window' or 'token_bucket'")

    async def check(self, key: str = "default") -> bool:
        """Check if request is allowed without incrementing counters"""
        if self.algorithm == "window":
            return await self._check_window(key)
        else:
            allowed, _ = await self.is_allowed(key)
            return allowed

    async def increment(self, key: str = "default") -> None:
        """Increment request counter for the given key"""
        if self.algorithm == "window":
            now = time()
            if key not in self._counters:
                self._counters[key] = {period: 0 for _, period in self.limits}
                self._last_reset[key] = now

            # Reset counters if period has elapsed
            if now - self._last_reset.get(key, 0) >= min(period for _, period in self.limits):
                self._counters[key] = {period: 0 for _, period in self.limits}
                self._last_reset[key] = now

            # Increment all counters
            for _, period in self.limits:
                self._counters[key][period] += 1

    async def _check_window(self, key: str = "default") -> bool:
        """Check if request is allowed using sliding window algorithm"""
        now = time()
        if key not in self._counters:
            return True

        # Check all limits
        for limit, period in self.limits:
            count = self._counters.get(key, {}).get(period, 0)
            if count >= limit and now - self._last_reset.get(key, 0) < period:
                return False
        return True

    async def is_allowed(self, key: str = "default") -> Tuple[bool, Dict[str, float]]:
        """
        Check if request is allowed and return detailed information
        
        Returns:
            Tuple[bool, Dict]: (allowed, info)
            info contains:
                - remaining: remaining tokens/requests
                - reset_after: seconds until reset
        """
        if self.algorithm == "token_bucket":
            success, wait_time = self._bucket.try_consume()
            remaining = self._bucket.get_tokens()
            
            return success, {
                "remaining": remaining,
                "reset_after": wait_time
            }
        else:
            # Window algorithm
            allowed = await self._check_window(key)
            if not allowed:
                # Calculate time until reset
                reset_after = min(
                    period - (time() - self._last_reset.get(key, 0))
                    for _, period in self.limits
                )
            else:
                reset_after = 0

            count = self._counters.get(key, {}).get(self.limits[0][1], 0)
            remaining = max(0, self.limits[0][0] - count)

            return allowed, {
                "remaining": remaining,
                "reset_after": max(0, reset_after)
            }



from typing import Callable, Dict, List, Optional
from enum import Enum

class RateLimitEvent(Enum):
    LIMIT_EXCEEDED = "limit_exceeded"
    NEAR_LIMIT = "near_limit"
    RESET = "reset"

class EventEmitter:
    def __init__(self):
        self.handlers: Dict[RateLimitEvent, List[Callable]] = {
            event: [] for event in RateLimitEvent
        }
        
    def on(self, event: RateLimitEvent, handler: Callable):
        """Register an event handler"""
        self.handlers[event].append(handler)
        
    def emit(self, event: RateLimitEvent, **kwargs):
        """Emit an event to all registered handlers"""
        for handler in self.handlers[event]:
            handler(**kwargs) 


class RateLimitExceeded(Exception):
    """Raised when rate limit is exceeded"""
    def __init__(self, message="Rate limit exceeded", retry_after=None):
        self.retry_after = retry_after
        super().__init__(message)


class StorageError(Exception):
    """Raised when storage backend encounters an error"""
    pass 


## tested complete 

import pytest
from ratehawk.monitoring.metrics import RateLimitMetrics

@pytest.fixture
def rate_limit_metrics():
    return RateLimitMetrics()

@pytest.mark.asyncio
async def test_track_request(rate_limit_metrics):
    key = "test_key"
    rate_limit_metrics.track_request(key, True)
    assert True  # Add appropriate assertions based on your metrics implementation

@pytest.mark.asyncio
async def test_update_usage(rate_limit_metrics):
    key = "test_key"
    rate_limit_metrics.update_usage(key, 5, 10)
    assert True  # Add appropriate assertions based on your metrics implementation

@pytest.mark.asyncio
async def test_track_exceeded(rate_limit_metrics):
    key = "test_key"
    rate_limit_metrics.track_exceeded(key)
    assert True  # Add appropriate assertions based on your metrics implementation

@pytest.mark.asyncio
async def test_start_timing(rate_limit_metrics):
    key = "test_key"
    rate_limit_metrics.start_timing(key)
    assert True  # Add appropriate assertions based on your metrics implementation

@pytest.mark.asyncio
async def test_stop_timing(rate_limit_metrics):
    key = "test_key"
    rate_limit_metrics.start_timing("test_key")
    rate_limit_metrics.stop_timing("test_key")
    assert True  # Add appropriate assertions based on your metrics implementation



from abc import ABC, abstractmethod
from typing import Tuple

class BaseStorage(ABC):
    """Base storage interface for rate limiting"""
    
    @abstractmethod
    async def increment(self, key: str, expiry: int) -> int:
        """Increment the counter for the given key"""
        pass
    
    @abstractmethod
    async def get(self, key: str) -> int:
        """Get the current count for the given key"""
        pass
    
    @abstractmethod
    async def reset(self, key: str) -> None:
        """Reset the counter for the given key"""
        pass

    @abstractmethod
    async def get_tokens(self, key: str) -> Tuple[int, float]:
        """Get the current tokens and last update time for the given key"""
        pass

    @abstractmethod
    async def set_tokens(self, key: str, tokens: int, timestamp: float) -> None:
        """Set the tokens and last update time for the given key"""
        pass

    @abstractmethod
    async def distributed_lock(self, key: str, timeout: int = 1):
        """Acquire a distributed lock for the given key"""
        pass



from setuptools import setup, find_packages

setup(
    name="ratehawk",
    version="0.1.6",
    packages=find_packages(include=['ratehawk', 'ratehawk.*']),
    install_requires=[
        "redis>=4.0.0",
        "aiosqlite>=0.17.0",
        "asyncpg>=0.27.0",
        "prometheus-client>=0.12.0",
        "python-dotenv>=0.19.0",
    ],
    extras_require={
        "dev": [
            "pytest>=6.0.0",
            "pytest-asyncio>=0.14.0",
            "black>=20.8b1",
            "isort>=5.6.4",
        ],
        "monitoring": [
            "prometheus-client>=0.12.0",
            "grafana-api-client>=0.3.4",
            "python-dateutil>=2.8.2",
        ],
    },
    author="Andrew Wade",
    author_email="hi@wadedev.online",
    description="A flexible API rate limiting library for Python",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/wadedesign/ratehawk",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    python_requires=">=3.7",
)



from .events import EventEmitter, RateLimitEvent
from .metrics import RateLimitMetrics
from .logging import RateLimitLogger

__all__ = [
    "EventEmitter",
    "RateLimitEvent",
    "RateLimitMetrics",
    "RateLimitLogger",
] 


import asyncio
import time
from typing import Optional, Tuple
import aiosqlite
from contextlib import asynccontextmanager

from ..exceptions import StorageError
from .base import BaseStorage


class SQLiteStorage(BaseStorage):
    def __init__(self, db_path: str = "ratehawk.db"):
        self.db_path = db_path
        self._db: Optional[aiosqlite.Connection] = None
        self._lock = asyncio.Lock()

    async def _ensure_connection(self):
        if self._db is None:
            try:
                self._db = await aiosqlite.connect(self.db_path)
                await self._db.execute("""
                    CREATE TABLE IF NOT EXISTS rate_limits (
                        key TEXT PRIMARY KEY,
                        tokens INTEGER NOT NULL,
                        last_update REAL NOT NULL
                    )
                """)
                await self._db.commit()
            except Exception as e:
                raise StorageError(f"Failed to connect to SQLite database: {str(e)}")

    async def get_tokens(self, key: str) -> Tuple[int, float]:
        await self._ensure_connection()
        async with self._lock:
            try:
                async with self._db.execute(
                    "SELECT tokens, last_update FROM rate_limits WHERE key = ?",
                    (key,)
                ) as cursor:
                    row = await cursor.fetchone()
                    if row is None:
                        return 0, 0.0
                    return row[0], row[1]
            except Exception as e:
                raise StorageError(f"Failed to get tokens from SQLite: {str(e)}")

    async def set_tokens(self, key: str, tokens: int, timestamp: float) -> None:
        await self._ensure_connection()
        async with self._lock:
            try:
                await self._db.execute("""
                    INSERT INTO rate_limits (key, tokens, last_update)
                    VALUES (?, ?, ?)
                    ON CONFLICT(key) DO UPDATE SET
                        tokens = excluded.tokens,
                        last_update = excluded.last_update
                """, (key, tokens, timestamp))
                await self._db.commit()
            except Exception as e:
                raise StorageError(f"Failed to set tokens in SQLite: {str(e)}")

    async def reset(self, key: str) -> None:
        await self._ensure_connection()
        async with self._lock:
            try:
                await self._db.execute("DELETE FROM rate_limits WHERE key = ?", (key,))
                await self._db.commit()
            except Exception as e:
                raise StorageError(f"Failed to reset tokens in SQLite: {str(e)}")

    async def close(self) -> None:
        if self._db is not None:
            try:
                await self._db.close()
                self._db = None
            except Exception as e:
                raise StorageError(f"Failed to close SQLite connection: {str(e)}")

    async def increment(self, key: str, expiry: int) -> int:
        await self._ensure_connection()
        async with self._lock:
            try:
                current_time = time.time()
                await self._db.execute("""
                    DELETE FROM rate_limits WHERE key = ? AND ? - last_update >= ?
                """, (key, current_time, expiry))
                await self._db.execute("""
                    INSERT INTO rate_limits (key, tokens, last_update)
                    VALUES (?, 1, ?)
                    ON CONFLICT(key) DO UPDATE SET
                        tokens = rate_limits.tokens + 1,
                        last_update = excluded.last_update
                """, (key, current_time))
                await self._db.commit()
                async with self._db.execute(
                    "SELECT tokens FROM rate_limits WHERE key = ?",
                    (key,)
                ) as cursor:
                    row = await cursor.fetchone()
                    return row[0] if row else 0
            except Exception as e:
                raise StorageError(f"Failed to increment tokens in SQLite: {str(e)}")

    async def get(self, key: str) -> int:
        await self._ensure_connection()
        async with self._lock:
            try:
                async with self._db.execute(
                    "SELECT tokens FROM rate_limits WHERE key = ?",
                    (key,)
                ) as cursor:
                    row = await cursor.fetchone()
                    return row[0] if row else 0
            except Exception as e:
                raise StorageError(f"Failed to get tokens from SQLite: {str(e)}")

    @asynccontextmanager
    async def distributed_lock(self, key: str, timeout: int = 1):
        """
        A simple distributed lock implementation using SQLite.
        While not truly distributed, it provides basic locking functionality.
        """
        lock_key = f"lock_{key}"
        try:
            await self._ensure_connection()
            async with self._lock:
                # Try to acquire lock
                await self._db.execute("""
                    CREATE TABLE IF NOT EXISTS locks (
                        lock_key TEXT PRIMARY KEY,
                        acquired_at REAL NOT NULL
                    )
                """)
                
                # Clean up old locks
                current_time = time.time()
                await self._db.execute(
                    "DELETE FROM locks WHERE ? - acquired_at >= ?",
                    (current_time, timeout)
                )
                
                # Try to acquire the lock
                try:
                    await self._db.execute(
                        "INSERT INTO locks (lock_key, acquired_at) VALUES (?, ?)",
                        (lock_key, current_time)
                    )
                    await self._db.commit()
                    yield
                except aiosqlite.IntegrityError:
                    # Lock is already held
                    raise StorageError("Failed to acquire lock")
        finally:
            # Release the lock
            async with self._lock:
                await self._db.execute("DELETE FROM locks WHERE lock_key = ?", (lock_key,))
                await self._db.commit()
